package org.openntf.wf.xsp.providers;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import org.openntf.wf.kernel.IWorkflowNotificationProvider;
import org.openntf.wf.kernel.WorkflowNotfication;
import org.openntf.wf.kernel.WorkflowRuntimeException;
import org.openntf.wf.xsp.util.Utils;

import com.ibm.xsp.extlib.util.ExtLibUtil;

import lotus.domino.Database;
import lotus.domino.Document;
import lotus.domino.DocumentCollection;
import lotus.domino.NotesException;
import lotus.domino.RichTextItem;
import lotus.domino.View;

public abstract class AbstractNotesNotificationStorageProvider implements IWorkflowNotificationProvider {

	private static final String VIEW_NAME = "lupWFMailTemplateByWFStepId";

	@Override
	public List<WorkflowNotfication> getAllByStepRuleId(String ruleId) {
		List<WorkflowNotfication> wfNotificationList = new ArrayList<WorkflowNotfication>();
		Database notificationDb = null;
		View viwLUP = null;
		try {
			notificationDb = getNotificationDatabase();
			viwLUP = notificationDb.getView(VIEW_NAME);
			DocumentCollection notificationCollection = viwLUP.getAllDocumentsByKey(ruleId,true);
			Document docNext = notificationCollection.getFirstDocument();
			while (docNext != null) {
				Document docProcess = docNext;
				docNext = notificationCollection.getNextDocument();
				WorkflowNotfication notification = buildNotification(docProcess);
				wfNotificationList.add(notification);
			}
		} catch (NotesException ex) {
			throw new WorkflowRuntimeException("Error during get all notifications", ex);
		}
		return wfNotificationList;
	}

	@SuppressWarnings("unchecked")
	private WorkflowNotfication buildNotification(Document docProcess) throws NotesException {
		WorkflowNotfication wfNotification = new WorkflowNotfication();
		wfNotification.setMessage(docProcess.getItemValueString("Body"));
		wfNotification.setRecipientCalculated("1".equals(docProcess.getItemValueString("recipientCalculated")));
		wfNotification.setRecipientExpression(docProcess.getItemValueString("recipientExpression"));
		wfNotification.setRecipients(docProcess.getItemValue("Recipients"));
		wfNotification.setSubject(docProcess.getItemValueString("Body"));
		return wfNotification;
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public void sendMessage(List<String> recipients, String subject, String body) {
		Document docMemo = null;
		RichTextItem bodyRT = null;
		try {
			Database ndbCurrent = ExtLibUtil.getCurrentDatabase();
			docMemo = ndbCurrent.createDocument();
			docMemo.replaceItemValue("Form", "memo");
			docMemo.replaceItemValue("Subject", subject);
			bodyRT = docMemo.createRichTextItem("body");
			bodyRT.appendText(body);
			docMemo.send(false,new Vector( recipients));
		} catch (NotesException ex) {
			throw new WorkflowRuntimeException("Error during send notifications", ex);
		} finally {
			Utils.recyler(docMemo, bodyRT);
		}
	}
	
	protected abstract Database getNotificationDatabase() throws NotesException;

}
