package org.openntf.wf.xsp.providers;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Vector;

import org.openntf.wf.kernel.IWorkflowProtocolStore;
import org.openntf.wf.kernel.WorkflowObject;
import org.openntf.wf.kernel.WorkflowProtocolEntry;
import org.openntf.wf.kernel.WorkflowRuntimeException;
import org.openntf.wf.xsp.util.Utils;

import lotus.domino.Database;
import lotus.domino.DateTime;
import lotus.domino.Document;
import lotus.domino.DocumentCollection;
import lotus.domino.NotesException;
import lotus.domino.View;

public abstract class AbstractNotesProtocolStorageProvider implements IWorkflowProtocolStore {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private static final String FIELD_ACTIVE = "active";
	private static final String FIELD_START_DATE = "startDate";
	private static final String FIELD_PROTOCOL_ID = "protocolId";
	private static final String FIELD_ID = "ID";
	private static final String FIELD_WORKFLOW_END_DATE = "WorkFlowEndDate";
	private static final String FIELD_WORKFLOW_START_DATE = "WorkFlowStartDate";
	private static final String FIELD_STEP_START_DATE = "StepStartDate";
	private static final String PROTOCOL_ENTRY_FORM_VALUE = "wfProtocolEntry";
	private static final String PROTOCOL_FORM_VALUE = "wfObject";
	private static final String FIELD_APPLICATION_NAME = "applicationName";
	private static final String FIELD_STEP_NAME = "stepName";
	private static final String FIELD_WORKFLOW_ID = "workflowId";
	private static final String FIELD_USER_NAME = "userName";
	private static final String FIELD_STEP_ID_TO = "stepIdTo";
	private static final String FIELD_STEP_ID_FROM = "stepIdFrom";
	private static final String FIELD_RULE_ID = "ruleId";
	private static final String FIELD_RULE_EXPRESSION = "ruleExpression";
	private static final String FIELD_DATE = "date";
	private static final String FIELD_BO_ID = "businessObjectID";
	private static final String FIELD_FORM = "Form";
	private static final String FIELD_STEP_EDITORS = "stepEditors";
	private static final String FIELD_STEP_ID = "stepId";
	private static final String FIELD_APP_ID = "appId";
	public static final String PROTOCOL_VIEW_NAME = "lupWFObjectByAppIDBoIDwfID";
	private static final String PROTOCOLENTRIES_BYKEY_VIEW_NAME = "lupProtocolEntriesByAppIDBoIDWfID";
	private static final String PROTOCOLENTRIES_BYPROTOCLID_VIEW_NAME = "lupProtocolEntriesByProtocolID";

	@Override
	public boolean isInWorkflow(String appid, String boid, String workflowId) {
		NotesAccessProperties accessProperties = null;
		try {
			accessProperties = accessWfDocument(buildDocumentAccessKey(appid, boid, workflowId), null);
			if (accessProperties.getWfDocument() == null) {
				return false;
			}
		} catch (NotesException e) {
			throw new WorkflowRuntimeException("General Error during load (" + appid + " / " + boid + " / " + workflowId + ")", e);

		} finally {
			if (accessProperties != null) {
				Utils.recyler(accessProperties.getAll4Recycle());
			}
		}
		return true;
	}

	@Override
	public WorkflowObject create(String appid, String boid, String workflowId, Serializable bo) {
		WorkflowObject wfObject = new WorkflowObject();
		wfObject.setBusinessObject(bo);
		wfObject.setAppId(appid);
		wfObject.setBoId(boid);
		wfObject.setWorkflowId(workflowId);
		return wfObject;
	}

	@SuppressWarnings("unchecked")
	@Override
	public WorkflowObject load(String appid, String boid, String workflowId, Serializable bo) {
		NotesAccessProperties accessProperties = null;
		try {
			accessProperties = accessWfDocument(buildDocumentAccessKey(appid, boid, workflowId), null);
			if (accessProperties.getWfDocument() == null) {
				throw new WorkflowRuntimeException("No Workflow Protocol found for " + appid + " / " + boid, null);
			}
			WorkflowObject wfObject = new WorkflowObject();
			wfObject.setBusinessObject(bo);
			wfObject.setId(accessProperties.getItemValueString(FIELD_ID));
			wfObject.setAppId(accessProperties.getWfDocument().getItemValueString(FIELD_APP_ID));
			wfObject.setBoId(accessProperties.getWfDocument().getItemValueString(FIELD_BO_ID));
			wfObject.setStepId(accessProperties.getWfDocument().getItemValueString(FIELD_STEP_ID));
			wfObject.setWorkflowId(accessProperties.getWfDocument().getItemValueString(FIELD_WORKFLOW_ID));
			wfObject.setStepEditors(accessProperties.getWfDocument().getItemValue(FIELD_STEP_EDITORS));
			wfObject.setStepName(accessProperties.getItemValueString(FIELD_STEP_NAME));
			wfObject.setApplicationName(accessProperties.getItemValueString(FIELD_APPLICATION_NAME));
			wfObject.setStepStartDate(getDateFormDocument(accessProperties.getWfDocument(), FIELD_STEP_START_DATE));
			wfObject.setWorkflowStartDate(getDateFormDocument(accessProperties.getWfDocument(), FIELD_WORKFLOW_START_DATE));
			wfObject.setWorkflowEndDate(getDateFormDocument(accessProperties.getWfDocument(), FIELD_WORKFLOW_END_DATE));
			wfObject.setActive("1".equals(accessProperties.getItemValueString(FIELD_ACTIVE)));
			return wfObject;
		} catch (NotesException e) {
			throw new WorkflowRuntimeException("General Error during load (" + appid + " / " + boid + ")", e);

		} finally {
			if (accessProperties != null) {
				Utils.recyler(accessProperties.getAll4Recycle());
			}
		}
	}

	private Date getDateFormDocument(Document doc, String fieldName) throws NotesException {
		if (!doc.hasItem(fieldName)) {
			return null;
		}
		@SuppressWarnings("rawtypes")
		Vector values = doc.getItemValueDateTimeArray(fieldName);
		if (values.isEmpty() || !(values.firstElement() instanceof DateTime)) {
			return null;
		}
		DateTime dtCurrent = (DateTime) values.firstElement();
		Date rcDate = dtCurrent.toJavaDate();
		dtCurrent.recycle();
		return rcDate;
	}

	private void setDateToDocument(Document doc, Date date, String fieldName) throws NotesException {
		if (date == null && doc.hasItem(fieldName)) {
			doc.removeItem(fieldName);
		}
		if (date == null) {
			return;
		}
		DateTime dtDate = doc.getParentDatabase().getParent().createDateTime(date);
		doc.replaceItemValue(fieldName, dtDate);
		dtDate.recycle();
	}

	public abstract Database getProtocolStoreDatabase() throws NotesException;

	private String buildDocumentAccessKey(String appid, String boid, String workflowId) {
		return appid + "@@@" + boid + "@@@" + workflowId;
	}

	@Override
	public void save(WorkflowObject wfObject) {
		NotesAccessProperties accessProperties = null;
		try {
			accessProperties = accessWfDocument(buildDocumentAccessKey(wfObject.getAppId(), wfObject.getBoId(), wfObject.getWorkflowId()), wfObject);
			if (accessProperties.getWfDocument() == null) {
				accessProperties.createNewDocument();
				accessProperties.getWfDocument().replaceItemValue(FIELD_FORM, PROTOCOL_FORM_VALUE);
				accessProperties.getWfDocument().replaceItemValue(FIELD_ID, wfObject.getId());
				accessProperties.getWfDocument().replaceItemValue(FIELD_APP_ID, wfObject.getAppId());
				accessProperties.getWfDocument().replaceItemValue(FIELD_BO_ID, wfObject.getBoId());
			}
			accessProperties.getWfDocument().replaceItemValue(FIELD_STEP_ID, wfObject.getStepId());
			accessProperties.getWfDocument().replaceItemValue(FIELD_WORKFLOW_ID, wfObject.getWorkflowId());
			accessProperties.getWfDocument().replaceItemValue(FIELD_STEP_EDITORS, new Vector<String>(wfObject.getStepEditors()));
			accessProperties.getWfDocument().replaceItemValue(FIELD_STEP_NAME, wfObject.getStepName());
			accessProperties.getWfDocument().replaceItemValue(FIELD_APPLICATION_NAME, wfObject.getApplicationName());
			setDateToDocument(accessProperties.getWfDocument(), wfObject.getStepStartDate(), FIELD_STEP_START_DATE);
			setDateToDocument(accessProperties.getWfDocument(), wfObject.getWorkflowStartDate(), FIELD_WORKFLOW_START_DATE);
			setDateToDocument(accessProperties.getWfDocument(), wfObject.getWorkflowEndDate(), FIELD_WORKFLOW_END_DATE);
			accessProperties.getWfDocument().replaceItemValue(FIELD_ACTIVE, wfObject.isActive() ? "1" : "");
			accessProperties.getWfDocument().save();
		} catch (NotesException e) {
			throw new WorkflowRuntimeException("General Error during save", wfObject, e);

		} finally {
			if (accessProperties != null) {
				Utils.recyler(accessProperties.getAll4Recycle());
			}
		}
	}

	private NotesAccessProperties accessWfDocument(String accessKey, WorkflowObject wfObject) throws NotesException {
		NotesAccessProperties accessProperties;
		accessProperties = new NotesAccessProperties(getProtocolStoreDatabase());
		if (accessProperties.getWfDatabase() == null) {
			throw new WorkflowRuntimeException("No Protocolstore defined", wfObject, null);
		}
		accessProperties.accessView(getProtocolViewName());
		accessProperties.accessDocument(accessKey);
		return accessProperties;
	}

	protected String getProtocolViewName() {
		return PROTOCOL_VIEW_NAME;
	}

	@Override
	public void saveEntry(WorkflowProtocolEntry entry) {
		Database ndbProtocol = null;
		Document docEntry = null;
		try {
			ndbProtocol = getProtocolStoreDatabase();
			docEntry = ndbProtocol.createDocument();
			docEntry.replaceItemValue(FIELD_FORM, PROTOCOL_ENTRY_FORM_VALUE);
			docEntry.replaceItemValue(FIELD_ID, entry.getId());
			docEntry.replaceItemValue(FIELD_PROTOCOL_ID, entry.getProtoclId());
			docEntry.replaceItemValue(FIELD_APP_ID, entry.getAppId());
			docEntry.replaceItemValue(FIELD_BO_ID, entry.getBoId());
			setDateToDocument(docEntry, entry.getDate(), FIELD_DATE);
			setDateToDocument(docEntry, entry.getStartDate(), FIELD_START_DATE);
			docEntry.replaceItemValue(FIELD_RULE_EXPRESSION, entry.getRuleExpression());
			docEntry.replaceItemValue(FIELD_RULE_ID, entry.getRuleId());
			docEntry.replaceItemValue(FIELD_STEP_ID_FROM, entry.getStepIdFrom());
			docEntry.replaceItemValue(FIELD_STEP_ID_TO, entry.getStepIdTo());
			docEntry.replaceItemValue(FIELD_USER_NAME, entry.getUserName());
			docEntry.replaceItemValue(FIELD_WORKFLOW_ID, entry.getWorkflowId());
			docEntry.save(true, false, true);
		} catch (NotesException ex) {
			throw new WorkflowRuntimeException("Error during save of WorkflowProtocolEntry", ex);
		} finally {
			Utils.recyler(ndbProtocol, docEntry);
		}
	}

	@Override
	public List<WorkflowProtocolEntry> getAllProtocolEntries(String appid, String boId, String workflowId) {
		List<WorkflowProtocolEntry> entries = new ArrayList<WorkflowProtocolEntry>();
		Database ndbProtocol = null;
		View viewEntries = null;
		DocumentCollection collection = null;
		Document docNext = null;

		try {
			ndbProtocol = getProtocolStoreDatabase();
			viewEntries = ndbProtocol.getView(PROTOCOLENTRIES_BYKEY_VIEW_NAME);
			collection = viewEntries.getAllDocumentsByKey(appid + "@@@" + boId + "@@@" + workflowId, true);
			docNext = collection.getFirstDocument();
			while (docNext != null) {
				Document docProcess = docNext;
				docNext = collection.getNextDocument();
				WorkflowProtocolEntry wfPEntry = buildWFProtocolEntry(docProcess);
				entries.add(wfPEntry);
				docProcess.recycle();
			}
		} catch (NotesException e) {
			throw new WorkflowRuntimeException("Error during getAllProtocolEntries", e);
		} finally {
			Utils.recyler(ndbProtocol, viewEntries, collection, docNext);
		}

		return entries;
	}

	private WorkflowProtocolEntry buildWFProtocolEntry(Document docEntry) throws NotesException {
		WorkflowProtocolEntry wfPEntry = new WorkflowProtocolEntry();
		wfPEntry.setId(docEntry.getItemValueString(FIELD_ID));
		wfPEntry.setProtoclId(docEntry.getItemValueString(FIELD_PROTOCOL_ID));
		wfPEntry.setAppId(docEntry.getItemValueString(FIELD_APP_ID));
		wfPEntry.setBoId(docEntry.getItemValueString(FIELD_BO_ID));
		wfPEntry.setDate(getDateFormDocument(docEntry, FIELD_DATE));
		wfPEntry.setStartDate(getDateFormDocument(docEntry, FIELD_START_DATE));
		wfPEntry.setRuleExpression(docEntry.getItemValueString(FIELD_RULE_EXPRESSION));
		wfPEntry.setRuleId(docEntry.getItemValueString(FIELD_RULE_ID));
		wfPEntry.setStepIdFrom(docEntry.getItemValueString(FIELD_STEP_ID_FROM));
		wfPEntry.setStepIdTo(docEntry.getItemValueString(FIELD_STEP_ID_TO));
		wfPEntry.setUserName(docEntry.getItemValueString(FIELD_USER_NAME));
		wfPEntry.setWorkflowId(docEntry.getItemValueString(FIELD_WORKFLOW_ID));
		return wfPEntry;
	}

	@Override
	public List<WorkflowProtocolEntry> getAllProtocolEntriesByProtocolID(String protocolId) {
		List<WorkflowProtocolEntry> entries = new ArrayList<WorkflowProtocolEntry>();
		Database ndbProtocol = null;
		View viewEntries = null;
		DocumentCollection collection = null;
		Document docNext = null;

		try {
			ndbProtocol = getProtocolStoreDatabase();
			viewEntries = ndbProtocol.getView(PROTOCOLENTRIES_BYPROTOCLID_VIEW_NAME);
			collection = viewEntries.getAllDocumentsByKey(protocolId, true);
			docNext = collection.getFirstDocument();
			while (docNext != null) {
				Document docProcess = docNext;
				docNext = collection.getNextDocument();
				WorkflowProtocolEntry wfPEntry = buildWFProtocolEntry(docProcess);
				entries.add(wfPEntry);
				docProcess.recycle();
			}
		} catch (NotesException e) {
			throw new WorkflowRuntimeException("Error during getAllProtocolEntries", e);
		} finally {
			Utils.recyler(ndbProtocol, viewEntries, collection, docNext);
		}

		return entries;
	}
}
