package org.openntf.wf.xsp.providers;

import java.util.ArrayList;
import java.util.List;

import javax.faces.FacesException;
import javax.faces.context.FacesContext;
import javax.faces.el.MethodBinding;

import org.openntf.wf.kernel.WorkflowObject;
import org.openntf.wf.kernel.rules.IRuleEvaluator;

import com.ibm.xsp.binding.MethodBindingEx;

public class ELRuleEvaluator implements IRuleEvaluator {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private class ParameterContainer {
		List<Class<? extends Object>> classes = new ArrayList<Class<? extends Object>>();
		List<String> variableNames = new ArrayList<String>();
		List<Object> objects = new ArrayList<Object>();

		private ParameterContainer(String varName, Object object) {
			add(varName, object);
		}

		public void add(String varName, Object object) {
			classes.add(object.getClass());
			variableNames.add(varName);
			objects.add(object);
		}

		public String[] getParameters() {
			return variableNames.toArray(new String[variableNames.size()]);
		}

		@SuppressWarnings("rawtypes")
		public Class[] getClasses() {
			return classes.toArray(new Class[classes.size()]);
		}

		public Object[] getObjects() {
			return objects.toArray();
		}

	}

	@Override
	public boolean isValid(String expression, Object businessObject) {
		ParameterContainer container = new ParameterContainer("bo", businessObject);
		return executeMethodBinding(expression, container, Boolean.class);
	}

	@SuppressWarnings("unchecked")
	@Override
	public List<String> getNotificationRecipient(String expression, Object businessObject) {
		ParameterContainer container = new ParameterContainer("bo", businessObject);
		return executeMethodBinding(expression, container, List.class);
	}

	@SuppressWarnings("unchecked")
	@Override
	public List<String> getStepResponsible(String expression, Object businessObject) {
		ParameterContainer container = new ParameterContainer("bo", businessObject);
		return executeMethodBinding(expression, container, List.class);
	}

	@Override
	public String getVariableValue(String expression, Object businessObject, WorkflowObject wfObject) {
		ParameterContainer container = new ParameterContainer("bo", businessObject);
		container.add("wf", wfObject);
		return executeMethodBinding(expression, container, String.class);
	}
	@Override
	public void executeOperation(String operation, Object businessObject, WorkflowObject wfObject) {
		ParameterContainer container = new ParameterContainer("bo", businessObject);
		container.add("wf", wfObject);
		invokeSSJS(operation, container);
	}
	
	@SuppressWarnings("unchecked")
	private <T> T executeMethodBinding(String expression, ParameterContainer container, Class<T> clazz) {
		Object obj = invokeSSJS(expression, container);
		if (obj == null) {
			throw new FacesException("Return value schould not be null");
		}
		if (obj.getClass().equals(String.class) && clazz.equals(List.class)) {
			List<String> lstString = new ArrayList<String>();
			lstString.add((String) obj);
			return (T) lstString;
		}
		if (obj.getClass().equals(clazz)) {
			return (T) obj;
		} else {
			throw new FacesException("Return value schould be type of " + clazz.getCanonicalName() + " and not "
					+ obj.getClass().getCanonicalName());
		}
	}

	private Object invokeSSJS(String expression, ParameterContainer container) {
		Object obj = null;
		MethodBinding mb = FacesContext.getCurrentInstance().getApplication().createMethodBinding("#{javascript:"+expression+"}",
				container.getClasses());
		if (mb instanceof MethodBindingEx) {
			MethodBindingEx mbEx = (MethodBindingEx) mb;
			mbEx.setParamNames(container.getParameters());
		}
		try {

			obj = mb.invoke(FacesContext.getCurrentInstance(), container.getObjects());
		} catch (Exception ex) {
			throw new FacesException("Error during Rule Evaluation - invoke", ex);
		}
		return obj;
	}
}
